//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include <kdrive_express.h>
#include <stdio.h>
#include <stdlib.h>

#define ERROR_MESSAGE_LEN  (128) /*!< kdriveExpress Error Messages */
#define MAX_USB_INTERFACES (10)  /*!< Max usb interfaces */

/*******************************
** Private Functions
********************************/

/*!
    Telegram Callback Handler
*/
static void on_telegram_callback(const uint8_t* telegram, uint32_t telegram_len, void* user_data);

/*!
    Secure Telegram Callback Handler
*/
static void on_sec_telegram_callback(const uint8_t* telegram, uint32_t telegram_len, void* user_data);

/*!
    Called when an error occurs
*/
static void error_callback(error_t e, void* user_data);

/*!
    Called when an event occurs
*/
static void event_callback(int32_t ap, uint32_t e, void* user_data);

/*******************************
** Private Variables
********************************/

/*
    Is the usb interface a RF-Interface.

    For Twisted Pair and Powerline:
        Set to 0

    For RF:
        Set to 1
*/
static bool_t is_rf = 0;

/*******************************
** Main
********************************/

int main(int argc, char* argv[])
{
	int32_t ap = 0;                                        /*! the Access Port descriptor */
	ip_tunn_dev_t iface_items[MAX_USB_INTERFACES];         /*!< holds the usb interface info found by USB enumeration */
	uint32_t iface_items_length  = MAX_USB_INTERFACES;     /*!< holds the number of items in iface_items */
	int32_t keyring              = 0;                      /*! the Keyring descriptor */
	const char* keyring_file     = "data/keyring.knxkeys"; /*! the path to keyring project file */
	const char* keyring_password = "123";                  /*! the password for the keyring project file */
	uint32_t key                 = 0;                      /*! used by the telegram callback mechanism. uniquely identifies a callback */
	int32_t sp                   = 0;                      /*! the Service Port descriptor */
	uint16_t individual_addr     = 0xAA1F;                 /*! the individual address of the secure device */
	uint16_t mask_version        = 0;                      /*! holds the mask version of the secure device */
	uint8_t data[255];                                     /*!< holds the data from the secure device*/
	uint32_t data_length = 255;                            /*!< holds the length of the data */
	uint16_t group_addr  = 0x2D06;                         /*!< The KNX Group Address (destination address) we use to send with */
	uint8_t value        = 1;                              /*!< The value we send on the bus with the Group Value Write command: on */

	/*
	    Configure the logging level and console logger
	*/
	kdrive_logger_set_level(KDRIVE_LOGGER_INFORMATION);
	kdrive_logger_console();

	/*
	    We register an error callback as a convenience logger function to
	    print out the error message when an error occurs.
	*/
	kdrive_register_error_callback(&error_callback, NULL);

	/*
	    We create a Access Port descriptor. This descriptor is then used for
	    all calls to that specific access port.
	*/
	ap = kdrive_ap_create();

	/*
	    We check that we were able to allocate a new descriptor
	    This should always happen, unless a bad_alloc exception is internally thrown
	    which means the memory couldn't be allocated.
	*/
	if (ap == KDRIVE_INVALID_DESCRIPTOR)
	{
		kdrive_logger(KDRIVE_LOGGER_FATAL, "Unable to create access port. This is a terminal failure");
		return -1;
	}

	/*
	    We register an event callback to notify of the Access Port Events
	    For example: KDRIVE_EVENT_TERMINATED
	*/
	kdrive_set_event_callback(ap, &event_callback, NULL);

	kdrive_ap_enum_ip_tunn(ap, iface_items, &iface_items_length);
	kdrive_logger_ex(KDRIVE_LOGGER_INFORMATION, "Found %d KNX IP Tunn Interfaces", iface_items_length);

	/*
	    Import the ETS keyring project
	*/
	if (kdrive_keyring_import_knxkeys(keyring_file, keyring_password, &keyring) != KDRIVE_ERROR_NONE)
	{
		kdrive_logger(KDRIVE_LOGGER_FATAL, "Unable to import the keyring file");
		return -1;
	}

	/* Use the keyring in the access port */
	kdrive_ap_set_keyring(ap, keyring);
	/* Send telegrams secure when the keyring contains a key */
	kdrive_ap_set_data_sec(ap, Auto);

	/* If we found at least 1 interface we simply open the first one (i.e. index 0) */
	if ((iface_items_length > 0) && kdrive_ap_open_ip_tunn(ap, "192.168.1.42", 0, NULL) == KDRIVE_ERROR_NONE)
	{
		/*
		    Connect the Packet Trace logging mechanism
		    to see the Rx and Tx packets
		*/
		kdrive_ap_packet_trace_connect(ap);

		/* now we simply go into bus monitor mode, and display received telegrams */
		kdrive_ap_register_telegram_callback(ap, &on_telegram_callback, NULL, &key);
		/* and received secure telegrams */
		kdrive_ap_register_sec_telegram_callback(ap, &on_sec_telegram_callback, NULL, &key);

		/* create a service port */
		sp = kdrive_sp_create(ap);

		/* read the device descriptor */
		kdrive_sp_device_descriptor_type0_read(sp, individual_addr, &mask_version);

		/* read the programming mode state */
		kdrive_sp_prop_value_read(sp, individual_addr, 0, 54, 1, 1, data, &data_length);

		/* send a 1-Bit boolean GroupValueWrite telegram */
		kdrive_ap_group_write(ap, group_addr, &value, 1);

		kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Entering BusMonitor Mode");
		kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Press [Enter] to exit the application ...");
		getchar();

		/* close the access port */
		kdrive_ap_close(ap);
	}

	/* releases the access port */
	kdrive_ap_release(ap);

	return 0;
}

/*******************************
** Private Functions
********************************/

/*!
    When a telegram is received we check to see if it is a L_Data.ind group value write
    telegram. If it is, we get the destination address and the datapoint value.
*/
void on_telegram_callback(const uint8_t* telegram, uint32_t telegram_len, void* user_data)
{
	kdrive_logger_dump(KDRIVE_LOGGER_INFORMATION, "Data :", telegram, telegram_len);
}

/*!
    When a secure telegram is received we check to see if it is a L_Data.ind group value
    write telegram. If it is, we get the destination address and the datapoint value.
*/
void on_sec_telegram_callback(const uint8_t* telegram, uint32_t telegram_len, void* user_data)
{
	kdrive_logger_dump(KDRIVE_LOGGER_INFORMATION, "SecData :", telegram, telegram_len);
}

/*!
    Called when a kdrive error exception is raised.
    The handling in the error callback is typically
    application specific. And here we simply show
    the error message.
*/
void error_callback(error_t e, void* user_data)
{
	if (e != KDRIVE_TIMEOUT_ERROR)
	{
		static char error_message[ERROR_MESSAGE_LEN];
		kdrive_get_error_message(e, error_message, ERROR_MESSAGE_LEN);
		kdrive_logger_ex(KDRIVE_LOGGER_ERROR, "kdrive error: %s", error_message);
	}
}

/*!
    The event callback is called when an Access Port event is raised
*/
void event_callback(int32_t ap, uint32_t e, void* user_data)
{
	switch (e)
	{
		case KDRIVE_EVENT_ERROR:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Access Port Error");
			break;

		case KDRIVE_EVENT_OPENING:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Access Port Opening");
			break;

		case KDRIVE_EVENT_OPENED:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Access Port Opened");
			break;

		case KDRIVE_EVENT_CLOSED:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Access Port Closed");
			break;

		case KDRIVE_EVENT_CLOSING:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Access Port Closing");
			break;

		case KDRIVE_EVENT_TERMINATED:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Access Port Terminated");
			break;

		case KDRIVE_EVENT_KNX_BUS_CONNECTED:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "KNX Bus Connected");
			break;

		case KDRIVE_EVENT_KNX_BUS_DISCONNECTED:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "KNX Bus Disconnected");
			break;

		case KDRIVE_EVENT_LOCAL_DEVICE_RESET:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Local Device Reset");
			break;

		case KDRIVE_EVENT_TELEGRAM_INDICATION:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Telegram Indication");
			break;

		case KDRIVE_EVENT_TELEGRAM_CONFIRM:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Telegram Confirm");
			break;

		case KDRIVE_EVENT_TELEGRAM_CONFIRM_TIMEOUT:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Telegram Confirm Timeout");
			break;

		case KDRIVE_EVENT_INTERNAL_01:
			break;

		default:
			kdrive_logger(KDRIVE_LOGGER_INFORMATION, "Unknown kdrive event");
			break;
	}
}
